@@ -5,6 +5,7 @@ Define MerakiToolKit class to ease operations with Meraki Cloud
 """
 
 # standard libraries
+import asyncio
 import os
 import sys
 import smtplib
@@ -17,12 +18,13 @@ from datetime import date
 
 # additional libraries
 import meraki
+import meraki.aio
 from . import merakitoolkitsupport
 
 __author__ = "Giovanni Augusto"
 __copyright__ = "Copyright (C) 2022 Giovanni Augusto"
 __license__ = "MIT"
-__version__ = "1.0.3"
+__version__ = "1.1.0"
 
 
 class MerakiToolkit():
@@ -50,7 +52,7 @@ class MerakiToolkit():
         self.apikey = settings["apikey"]
         # operation data received in input
         self.current_operation = settings
-        self.connect()
+        self.dashboard = None
 
 
     @property
@@ -130,14 +132,23 @@ class MerakiToolkit():
     def connect(self):
         '''
         connects to Meraki dashboard
+        ASYNC : returns an initialized meraki.aio.AsyncDashboardAPI object
+        object is to be used with 'async with' in calling methods (pskchangeasync)
+        and reference 'as' is to be self.dashboard
         '''
+        if self.current_operation["settings"]["verbose"] >= 3:
+            logging = True
+        else :
+            logging = False
         try:
-            self.dashboard = meraki.DashboardAPI(
+            return meraki.aio.AsyncDashboardAPI(
                 api_key=self.apikey,
-                suppress_logging=not self.current_operation["settings"]["verbose"],
+                suppress_logging=not logging,
                 simulate=False,
                 caller="merakitoolkit"
                 )
+        except meraki.exceptions.AsyncAPIError as err:
+            print(f'operation: {err.operation} error: {err.message["errors"]}')
         except meraki.exceptions.APIError as err:
             print(f'operation: {err.operation} error: {err.message["errors"]}')
         except Exception as err: # pylint: disable=broad-except
@@ -145,24 +156,51 @@ class MerakiToolkit():
             sys.exit(2)
 
 
-    def get_organizations(self):
+    async def get_organizations(self):
         '''
         Retrieve organizations from Meraki dashboard and return them
         '''
         try:
-            organizations = self.dashboard.organizations.getOrganizations()
+            if self.current_operation["settings"]["verbose"]>=2:
+                print("START: getting Organizations")
+            organizations = await self.dashboard.organizations.getOrganizations()
+            if self.current_operation["settings"]["verbose"]>=2:
+                print("END: getting Organizations")
             return organizations
+        except meraki.exceptions.AsyncAPIError as err:
+            # Too many requests
+            if err.response.status == 429:
+                # wait for the time indicated in reponse header Retry-After and then retry
+                await asyncio.sleep(int(err.response.headers["Retry-After"]))
+                return await self.get_organizations()
+            else:
+                print(f'operation: {err.operation} error: {err.message["errors"]}')
+                return None
         except Exception as err: # pylint: disable=broad-except
             print("An error occurred while retrieving Organizations: ",err)
             sys.exit(2)
 
-    def get_network_wireless_ssids(self,network):
+
+    async def get_network_wireless_ssids(self,network):
         '''
         Retrieve SSIDs from a Network in Meraki dashboard and return them
         '''
         try:
-            ssids = self.dashboard.wireless.getNetworkWirelessSsids(network["id"])
+            if self.current_operation["settings"]["verbose"]>=2:
+                print(f"START: getting SSIDs for Network: {network['name']}")
+            ssids = await self.dashboard.wireless.getNetworkWirelessSsids(network["id"])
+            if self.current_operation["settings"]["verbose"]>=2:
+                print(f"END: getting SSIDs for Network: {network['name']}")
             return ssids
+        except meraki.exceptions.AsyncAPIError as err:
+            # Too many requests
+            if err.response.status == 429:
+                # wait for the time indicated in reponse header Retry-After and then retry
+                await asyncio.sleep(int(err.response.headers["Retry-After"]))
+                return await self.get_network_wireless_ssids(network)
+            else:
+                print(f'operation: {err.operation} error: {err.message["errors"]} network: {network["name"]}')
+                return None
         except meraki.exceptions.APIError as err:
             print(f'operation: {err.operation} error: {err.message["errors"]} network: {network["name"]}')
             return None
@@ -170,13 +208,27 @@ class MerakiToolkit():
             print("An error occurred while retrieving Organizations: ",err)
             sys.exit(2)
 
-    def get_organization_networks(self,organization):
+
+    async def get_organization_networks(self,organization):
         '''
         Retrieve Networks from an organization in Meraki dashboard and return them
         '''
         try:
-            networks = self.dashboard.organizations.getOrganizationNetworks(organization["id"])
+            if self.current_operation["settings"]["verbose"]>=2:
+                print(f"START: getting networks for org: {organization['name']}")
+            networks = await self.dashboard.organizations.getOrganizationNetworks(organization["id"])
+            if self.current_operation["settings"]["verbose"]>=2:
+                print(f"END: getting networks for org: {organization['name']}")
             return networks
+        except meraki.exceptions.AsyncAPIError as err:
+            # Too many requests
+            if err.response.status == 429:
+                # wait for the time indicated in reponse header Retry-After and then retry
+                await asyncio.sleep(int(err.response.headers["Retry-After"]))
+                return await self.get_organization_networks(organization)
+            else:
+                print(f'operation: {err.operation} error: {err.message["errors"]} Organization: {organization["name"]}')
+                return None
         except meraki.exceptions.APIError as err:
             print(f'operation: {err.operation} error: {err.message["errors"]} Organization: {organization["name"]}')
             return None
@@ -184,16 +236,30 @@ class MerakiToolkit():
             print("An error occurred while retrieving Networks: ",err)
             sys.exit(2)
 
-    def update_network_wireless_ssid(self,network,passphrase):
+
+    async def update_network_wireless_ssid(self,network,passphrase):
         '''
         update Wireless SSID in a network and return outcome of the operation
         '''
         try:
-            ssid = self.dashboard.wireless.updateNetworkWirelessSsid(network["id"],network["ssidPosition"],psk=passphrase)
+            if self.current_operation["settings"]["verbose"]>=2:
+                print(f"START: updating PSK for network: {network['name']}")
+            ssid = await self.dashboard.wireless.updateNetworkWirelessSsid(network["id"],network["ssidPosition"],psk=passphrase)
+            if self.current_operation["settings"]["verbose"]>=2:
+                print(f"END: updating PSK for network: {network['name']}")
             if ssid["psk"] == passphrase:
                 return True
             else:
                 raise ValueError(f"PSK change : {ssid['name']} passhprase was not changed!")
+        except meraki.exceptions.AsyncAPIError as err:
+            # Too many requests
+            if err.response.status == 429:
+                # wait for the time indicated in reponse header Retry-After and then retry
+                await asyncio.sleep(int(err.response.headers["Retry-After"]))
+                return await self.update_network_wireless_ssid(network,passphrase)
+            else:
+                print(f'operation: {err.operation} error: {err.message["errors"]} Network: {network["id"]} SSID: {network["ssidName"]}') # pylint: disable=line-too-long
+                return False
         except meraki.exceptions.APIError as err:
             print(f'operation: {err.operation} error: {err.message["errors"]} Network: {network["id"]} SSID: {network["ssidName"]}') # pylint: disable=line-too-long
             return False
@@ -202,11 +268,43 @@ class MerakiToolkit():
             sys.exit(2)
 
 
-    def pskchange(self):
+    async def pskchangeasync(self):
         '''
         Change Pre Shared Key for an SSID in specified network name in organizations
         '''
+
+        # Coroutine to process Networks in an organization for PSK change
+        # returns data of a network to process (if parameters has a match)
+        # returns None if there is no match -> needs to explicitly cleanup later 'None' entries from networks list
+        async def process_network(organization,network,settings):
+            if (network["name"] not in settings["network"]) and ("ALL" not in settings["network"]):
+                return None
+            # verify that at least one of the TAGs is in the list of network tags
+            if settings["tags"]:
+                if not any( tag in settings["tags"] for tag in network["tags"]):
+                    return None
+            # retrieve SSIDs of the evaluated network
+            # uses awaitable method that will be leveraged later by asyncio.gather()
+            network_ssids = await self.get_network_wireless_ssids(network)
+            # some networks has no SSIDs (camera,appliance,etc) so we skip those
+            if not network_ssids:
+                return None
+            network_to_process = None
+            for ssidposition in range(len(network_ssids)): # pylint: disable=consider-using-enumerate
+                if settings["ssid"] == network_ssids[ssidposition]["name"]: # SSID is found
+                    # create dictionary to collect all necessary information
+                    network_to_process = {}
+                    network_to_process["organization"] = organization["name"]
+                    network_to_process["name"] = network["name"]
+                    network_to_process["id"] = network["id"]
+                    network_to_process["ssidPosition"] = str(ssidposition)
+                    network_to_process["ssidName"] = network_ssids[ssidposition]["name"]
+                    network_to_process["wpaEncryptionMode"] = network_ssids[ssidposition]["wpaEncryptionMode"]
+                    break # SSID was found -> exit the loop
+            return network_to_process
+
         settings = self.current_operation["settings"]
+
         try:
             if settings is None:
                 raise ValueError("PSK change : No operation has been defined")
@@ -224,59 +322,68 @@ class MerakiToolkit():
             # flag to set to save relevant data for other processes
             data_has_changed = False
 
-            # scan matching organizations -> networks -> ssid -> collect data
-            for organization in self.get_organizations():
-                # Verify that the current organization is in the list of organizations to process
-                if organization["name"] in settings["organization"] or "ALL" in settings["organization"]:
-                    # Retrieve Networks for the current organization
-                    networks = self.get_organization_networks(organization)
-                    for network in networks:
-                        # Verify that network name is a match, if not skip this cycle
-                        if (network["name"] not in settings["network"]) and ("ALL" not in settings["network"]):
-                            continue
-                        # verify that at least one of the TAGs is in the list of network tags
-                        if settings["tags"]:
-                            if not any( tag in settings["tags"] for tag in network["tags"]):
-                                continue
-                        # retrieve SSIDs of the evaluated network
-                        network_ssids = self.get_network_wireless_ssids(network) # pylint: disable=line-too-long
-                        # some networks has no SSIDs (camera,appliance,etc) so we skip those
-                        if not network_ssids:
-                            continue
-                        for ssidposition in range(len(network_ssids)): # pylint: disable=consider-using-enumerate
-                            if settings["ssid"] == network_ssids[ssidposition]["name"]: # SSID is found
-                                # create dictionary to collect all necessary information
-                                network_to_process = {}
-                                network_to_process["organization"] = organization["name"]
-                                network_to_process["name"] = network["name"]
-                                network_to_process["id"] = network["id"]
-                                network_to_process["ssidPosition"] = str(ssidposition)
-                                network_to_process["ssidName"] = network_ssids[ssidposition]["name"]
-                                network_to_process["wpaEncryptionMode"] = network_ssids[ssidposition]["wpaEncryptionMode"]
-                                networks_to_process.append(network_to_process)
-                                break # SSID was found -> exit the loop
+            # Create context manager for the async mereaki.aio.AsyncDashboardAPI object (necessary to ensure a proper closure)
+            # Standard in MerakiToolKit is to store context manager variable in self.dashboard
+            # connect() method is used to aggregate all settings centrally
+            async with self.connect() as self.dashboard:
+                # async tasks collection:
+                # collect organizations->collect networks in organization->collect SSID->add item to list of networks to process
+
+                # async collect organizations (create task->await task)
+                task_organizations = asyncio.create_task(self.get_organizations())
+                organizations = await task_organizations
+                # process_networks_tasks -> list of coroutines for networks to process
+                process_networks_tasks = []
+                # in each organization retrieve list of networks
+                for organization in organizations:
+                    # Verify that the current organization is in the list of organizations to process
+                    if organization["name"] in settings["organization"] or "ALL" in settings["organization"]:
+                        # async collect list of networks (create task->await task)
+                        task_networks = asyncio.create_task(self.get_organization_networks(organization))
+                        networks = await task_networks
+                        # create a coroutine to process each network and add it to process_networks_tasks list
+                        for network in networks:
+                            process_networks_tasks.append(process_network(organization,network,settings))
+
+                # Collection happened through an iteration with async awaits
+                # here we process all coroutines collected asynchronously with asyncio.gather
+                # asyncio.gather -> returns values when all coroutines are completed
+                current_networks_to_process = await asyncio.gather(*process_networks_tasks)
+                # Clean current networks list of the null entries and keep only networks to process
+                # extend the final networks_to_process list with the interesting networks for the current org
+                networks_to_process.extend([x for x in current_networks_to_process if x is not None])
+
+
+
+                # Execution code : at this point data is being changed (or simulated) on Meraki Cloud
+                # NOTE : this portion could be extracted into a standalone executor method for multiple tasks
+                if settings["dryrun"] or settings["verbose"]>=1:
+                    if settings["dryrun"]:
+                        data_has_changed = True
+                        print("\033[91m","\nDRYRUN Enabled: Changes below will not be applied")
+                        print("\033[0m","-"*110)
+                    print(f'{"Organization:":<25} {"Network:":<45} {"SSID:":<20} {"PSK:":<20}')
+                    for network in networks_to_process:
+                        print("-"*110)
+                        print(f"{network['organization']:<25} {network['name']:<45} {network['ssidName']:<20} {settings['passphrase']:<20}") # pylint: disable=line-too-long
+                else:
+                    update_networks_tasks = []
+                    for network in networks_to_process:
+                        update_networks_tasks.append(self.update_network_wireless_ssid(network,settings["passphrase"]))
+                        #data_has_changed = self.update_network_wireless_ssid(network,settings["passphrase"])
+                    data_has_changed = True in await asyncio.gather(*update_networks_tasks)
+
+                # save last operation data only if a change (real or simulated) happened
+                if data_has_changed:
+                    self.current_operation["networks_to_process"] = networks_to_process
+                    self.current_operation["success"] = True
+
         except Exception as err: # pylint: disable=broad-except
             print("An error occurred while running PSK change: ",err)
             sys.exit(2)
 
-        # Execution code : at this point data is being changed (or simulated) on Meraki Cloud
-        # NOTE : this portion could be extracted into a standalone executor method for multiple tasks
-        if settings["dryrun"]:
-            print(f'{"Organization":<25} {"Network:":<45} {"SSID:":<20} {"PSK:":<20}')
-            for network in networks_to_process:
-                print("-"*100)
-                print(f"{network['organization']:<25} {network['name']:<45} {network['ssidName']:<20} {settings['passphrase']:<20}") # pylint: disable=line-too-long
-                data_has_changed = True
-        else:
-            for network in networks_to_process:
-                data_has_changed = self.update_network_wireless_ssid(network,settings["passphrase"])
 
-        # save last operation data only if a change (real or simulated) happened
-        if data_has_changed:
-            self.current_operation["networks_to_process"] = networks_to_process
-            self.current_operation["success"] = True
 
-    # refactor email send method
     def send_email_psk(self):
         ''' send email for PSK change notification'''
 
